from typing import Literal, Optional, Union

from browsy import BaseJob, Page


class PDFJob(BaseJob):
    """Generates a PDF from a webpage or HTML content."""

    NAME = "pdf"

    url: Optional[str] = None
    html: Optional[str] = None
    emulate_media: Union[Literal["null", "print", "screen"], None] = None

    async def execute(self, page: Page) -> bytes:
        if self.url:
            await page.goto(self.url)
        elif self.html:
            await page.set_content(self.html)

        if self.emulate_media:
            await page.emulate_media(self.emulate_media)

        return await page.pdf()

    async def validate_logic(self) -> bool:
        return bool(self.url) != bool(self.html)



import json
from datetime import datetime
from typing import Literal, Optional, Union, List, Tuple

import aiosqlite
from pydantic import BaseModel, field_validator

from browsy import _jobs

AsyncConnection = aiosqlite.Connection

_INIT_SQL = """
CREATE TABLE IF NOT EXISTS jobs (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL,
    input TEXT NOT NULL CHECK(json_valid(input)),
    status TEXT NOT NULL,
    created_at DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%f', 'now')),
    updated_at DATETIME,
    worker TEXT,
    processing_time INTEGER
);
CREATE INDEX IF NOT EXISTS idx_jobs_status ON jobs(status);

CREATE TABLE IF NOT EXISTS workers (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL UNIQUE,
    last_check_in_time DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%f', 'now')),
    last_activity_time DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%f', 'now'))
);
CREATE INDEX IF NOT EXISTS idx_workers_name ON workers(name);

CREATE TABLE IF NOT EXISTS outputs (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    job_id INTEGER NOT NULL,
    output BLOB,
    FOREIGN KEY (job_id) REFERENCES jobs (id)
);
CREATE INDEX IF NOT EXISTS idx_outputs_job_id ON outputs(job_id);
"""


class DBJobBase(BaseModel):
    id: int
    name: str
    status: _jobs.JobStatus
    created_at: datetime
    updated_at: Optional[datetime]
    worker: Optional[str]
    processing_time: Optional[int]  # milliseconds


class DBJob(DBJobBase):
    input: dict

    @field_validator("input", mode="before")
    @classmethod
    def json_str_output(cls, v: Union[str, dict]) -> dict:
        if isinstance(v, str):
            return json.loads(v)
        return v


class DBOutput(BaseModel):
    id: int
    job_id: int
    output: Optional[bytes]


class DBWorker(BaseModel):
    id: int
    name: str
    last_check_in_time: datetime
    last_activity_time: datetime

    @property
    def uptime(self) -> datetime:
        return datetime.now()


async def create_connection(db_path: str) -> AsyncConnection:
    conn = await aiosqlite.connect(db_path)
    conn.row_factory = aiosqlite.Row
    return conn


async def init_db(conn: AsyncConnection) -> None:
    await conn.execute("PRAGMA journal_mode = WAL;")
    await conn.commit()

    await conn.executescript(_INIT_SQL)
    await conn.commit()


async def create_job(
    conn: AsyncConnection,
    name: str,
    input_json: str,
) -> DBJob:
    async with conn.execute(
        """
        INSERT INTO jobs (name, input, status)
        VALUES (?, ?, ?)
        RETURNING id, created_at, updated_at, worker, processing_time
        """,
        (name, input_json, _jobs.JobStatus.PENDING),
    ) as cursor:
        result = await cursor.fetchone()

    await conn.commit()

    return DBJob(
        name=name,
        input=input_json,
        status=_jobs.JobStatus.PENDING,
        **result,
    )


async def get_job_by_id(
    conn: AsyncConnection,
    id_: int,
) -> Optional[DBJob]:
    async with conn.execute(
        """
        SELECT id, name, input, status, created_at, updated_at, worker, processing_time
        FROM jobs
        WHERE id = ?
        """,
        (id_,),
    ) as cursor:
        result = await cursor.fetchone()

    return DBJob(**result) if result else None


async def get_job_result_by_job_id(
    conn: AsyncConnection,
    job_id: int,
) -> Optional[bytes]:
    async with conn.execute(
        """
        SELECT id, job_id, output
        FROM outputs
        WHERE job_id = ?
        """,
        (job_id,),
    ) as cursor:
        result = await cursor.fetchone()

    return result["output"] if result else None


async def get_next_job(conn: AsyncConnection, worker: str) -> Optional[DBJob]:
    # Acquires a reserved lock, blocking other write transactions
    await conn.execute("BEGIN IMMEDIATE")

    async with conn.execute(
        f"""
        SELECT id, name, input, status, created_at, updated_at, worker, processing_time
        FROM jobs
        WHERE status = '{_jobs.JobStatus.PENDING.value}'
        ORDER BY created_at ASC
        LIMIT 1
        """
    ) as cursor:
        result = await cursor.fetchone()

    if not result:
        # Releases the lock
        await conn.rollback()
        return None

    db_job = DBJob(**result)

    await conn.execute(
        f"""
        UPDATE jobs
        SET status = '{_jobs.JobStatus.IN_PROGRESS.value}', updated_at = strftime('%Y-%m-%d %H:%M:%f', 'now'), worker = ?
        WHERE id = ?
        """,
        (worker, db_job.id),
    )

    await update_worker_activity(conn, worker, commit=False)

    await conn.commit()
    db_job.status = _jobs.JobStatus.IN_PROGRESS
    db_job.worker = worker

    return db_job


async def update_job_status(
    conn: AsyncConnection,
    worker: str,
    job_id: int,
    status: Literal[_jobs.JobStatus.DONE, _jobs.JobStatus.FAILED],
    processing_time: int,
    output: Optional[bytes],
) -> None:
    await conn.execute(
        """
        UPDATE jobs
        SET status = ?, updated_at = strftime('%Y-%m-%d %H:%M:%f', 'now'), processing_time = ?
        WHERE id = ?
        """,
        (status, processing_time, job_id),
    )

    if output:
        await conn.execute(
            """
            INSERT INTO outputs (job_id, output)
            VALUES (?, ?)
            """,
            (job_id, output),
        )

    await update_worker_activity(conn, worker, commit=False)

    await conn.commit()


async def check_in_worker(
    conn: AsyncConnection,
    worker: str,
) -> None:
    """Updates worker's last check-in timestamp."""
    await conn.execute(
        """
        INSERT INTO workers (name, last_check_in_time)
        VALUES (?, strftime('%Y-%m-%d %H:%M:%f', 'now'))
        ON CONFLICT(name) DO UPDATE SET
            last_check_in_time = strftime('%Y-%m-%d %H:%M:%f', 'now')
        """,
        (worker,),
    )
    await conn.commit()


async def update_worker_activity(
    conn: AsyncConnection, worker: str, commit: bool = True
) -> None:
    """Updates worker's last activity timestamp."""
    await conn.execute(
        """
        UPDATE workers
        SET last_activity_time = strftime('%Y-%m-%d %H:%M:%f', 'now')
        WHERE name = ?
        """,
        (worker,),
    )
    if commit:
        await conn.commit()


async def get_workers(
    conn: AsyncConnection, last_activity_time_ge: Optional[datetime] = None
) -> List[DBWorker]:
    args = []
    query = """SELECT id, name, last_check_in_time, last_activity_time
                    FROM workers"""
    if last_activity_time_ge:
        query += " WHERE last_activity_time >= ?"
        args.append(last_activity_time_ge)

    query += " ORDER BY last_activity_time DESC"

    async with conn.execute(query, args) as cursor:
        result = await cursor.fetchall()

    if not result:
        return []

    return [DBWorker(**r) for r in result]


async def get_jobs(
    conn: AsyncConnection,
    status: Optional[_jobs.JobStatus] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Tuple[List[DBJob], int]:
    limit = limit or 50
    offset = offset or 0

    args = []
    where_clause = ""
    if status:
        where_clause = " WHERE status = ? "
        args.append(status)

    count_query = f"SELECT COUNT(*) FROM jobs{where_clause}"
    async with conn.execute(count_query, args) as cursor:
        total_count = (await cursor.fetchone())[0]

    query = f"""
        SELECT id, name, status, created_at, updated_at, worker, processing_time
        FROM jobs{where_clause}
        ORDER BY created_at DESC
        LIMIT ? OFFSET ?
    """
    args.extend([limit, offset])

    async with conn.execute(query, args) as cursor:
        rows = await cursor.fetchall()

    jobs = [DBJobBase(**r) for r in rows]

    return jobs, total_count



import os
import importlib.resources as resources
from contextlib import asynccontextmanager
from typing import Annotated, Optional
from pathlib import Path
from datetime import datetime, timezone

from fastapi import FastAPI, Depends, HTTPException, Request, Response
from fastapi.responses import RedirectResponse
from fastapi.templating import Jinja2Templates
from fastapi.openapi.utils import get_openapi
from pydantic import BaseModel

from browsy import _database, _jobs, __version__
from browsy import __name__ as pkg_name

_JOBS_DEFS = _jobs.collect_jobs_defs(
    os.environ.get("BROWSY_JOBS_PATH", str(Path().absolute()))
)


def custom_openapi():
    """Custom OpenAPI to include jobs in schema that are dynamically loaded."""
    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title=app.title,
        version=app.version,
        description=app.description,
        routes=app.routes,
    )
    for _, job_cls in _JOBS_DEFS.items():
        openapi_schema["components"]["schemas"][
            job_cls.__name__
        ] = job_cls.model_json_schema()
    app.openapi_schema = openapi_schema
    return app.openapi_schema


@asynccontextmanager
async def lifespan(app: FastAPI):
    db_path = os.environ.get("BROWSY_DB_PATH")
    if not db_path:
        raise ValueError("BROWSY_DB_PATH not set")

    app.state.DB_PATH = db_path
    app.state.startup_time = datetime.now(timezone.utc)

    conn = await _database.create_connection(db_path)

    try:
        await _database.init_db(conn)
    finally:
        await conn.close()

    yield


app = FastAPI(
    lifespan=lifespan,
    title="browsy",
    version=__version__,
    redoc_url=None,
    openapi_tags=[
        {
            "name": "jobs",
            "description": (
                "Create, monitor, and retrieve results from browser"
                " automation jobs."
            ),
        }
    ],
)
app.openapi = custom_openapi

template_dir = resources.files(pkg_name) / "templates"
templates = Jinja2Templates(directory=template_dir)


async def get_db(request: Request):
    conn = await _database.create_connection(request.app.state.DB_PATH)

    try:
        yield conn
    finally:
        await conn.close()


class JobRequest(BaseModel):
    name: str
    parameters: dict


@app.post("/api/v1/jobs", response_model=_database.DBJob, tags=["jobs"])
async def submit_job(
    r: JobRequest,
    db_conn: Annotated[_database.AsyncConnection, Depends(get_db)],
):
    if r.name not in _JOBS_DEFS:
        raise HTTPException(400, "Job with that name is not defined.")

    job = _JOBS_DEFS[r.name].model_validate(r.parameters)
    is_valid = await job.validate_logic()
    if not is_valid:
        raise HTTPException(400, "Job validation failed")

    return await _database.create_job(db_conn, r.name, job.model_dump_json())


@app.get("/api/v1/jobs/{job_id}", response_model=_database.DBJob, tags=["jobs"])
async def get_job_by_id(
    job_id: int, db_conn: Annotated[_database.AsyncConnection, Depends(get_db)]
):
    job = await _database.get_job_by_id(db_conn, job_id)
    if not job:
        raise HTTPException(404, "Job not found")

    return job


@app.get("/api/v1/jobs/{job_id}/result", tags=["jobs"])
async def get_job_result_by_job_id(
    job_id: int, db_conn: Annotated[_database.AsyncConnection, Depends(get_db)]
):
    job = await _database.get_job_by_id(db_conn, job_id)
    if not job:
        raise HTTPException(404)

    headers = {
        "X-Job-Status": job.status,
        "X-Job-Last-Updated": job.updated_at.isoformat(),
    }

    if job.status in (_jobs.JobStatus.IN_PROGRESS, _jobs.JobStatus.PENDING):
        return Response(202, headers=headers)

    if job.status == _jobs.JobStatus.FAILED:
        return Response(204, headers=headers)

    job_result = await _database.get_job_result_by_job_id(db_conn, job_id)
    if job_result is None:
        return Response(204, headers=headers)

    return Response(
        content=job_result,
        status_code=200,
        media_type="application/octet-stream",
        headers=headers,
    )


@app.get("/health", include_in_schema=False)
async def healthcheck(_: Annotated[_database.AsyncConnection, Depends(get_db)]):
    return {"status": "ok", "version": __version__}


@app.get("/internal", include_in_schema=False)
async def get_monitoring():
    return RedirectResponse(url="/internal/workers")


@app.get("/internal/workers", include_in_schema=False)
async def get_workers_information(
    request: Request,
    db_conn: Annotated[_database.AsyncConnection, Depends(get_db)],
):
    workers = await _database.get_workers(
        db_conn, last_activity_time_ge=app.state.startup_time
    )
    return templates.TemplateResponse(
        request=request, name="workers.html", context={"workers": workers}
    )


@app.get("/internal/jobs", include_in_schema=False)
async def get_jobs_information(
    request: Request,
    db_conn: Annotated[_database.AsyncConnection, Depends(get_db)],
    limit: Optional[int] = None,
    offset: Optional[int] = None,
):
    jobs, count = await _database.get_jobs(db_conn, limit=limit, offset=offset)
    for j in jobs:
        j.status = j.status.value
    return templates.TemplateResponse(
        request=request,
        name="jobs.html",
        context={
            "jobs": jobs,
            "pagination": {"limit": limit, "offset": offset, "count": count},
        },
    )



import asyncio
import logging
import signal
import time
from typing import Optional

from playwright.async_api import (
    PlaywrightContextManager,
    Browser,
    Error as PlaywrightError,
)
from playwright._impl._errors import TargetClosedError

from browsy import _database, _jobs

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger("worker")

_JOB_POLL_INTERVAL = 5
_HEARTBEAT_LOG_INTERVAL = 60


async def _worker_loop(name: str, db_path: str, jobs_path: str) -> None:
    worker_logger = logging.getLogger(name)

    db = await _database.create_connection(db_path)
    jobs_defs = _jobs.collect_jobs_defs(jobs_path)
    await _database.check_in_worker(db, name)

    pw = await PlaywrightContextManager().start()
    browser: Browser = await pw.chromium.launch(headless=True)
    worker_logger.info("Browser launched and ready")

    worker_heartbeat = time.monotonic()
    job: Optional[_database.DBJob] = None
    start_time: Optional[float] = None

    try:
        while True:
            job = await _database.get_next_job(db, name)
            if not job:
                if (
                    time.monotonic() - worker_heartbeat
                    >= _HEARTBEAT_LOG_INTERVAL
                ):
                    await _database.update_worker_activity(db, name)
                    worker_heartbeat = time.monotonic()
                worker_logger.debug(
                    f"No jobs available, sleeping for {_JOB_POLL_INTERVAL}s"
                )
                await asyncio.sleep(_JOB_POLL_INTERVAL)
                continue

            worker_logger.info(f"Starting job {job.id} (type: {job.name})")

            start_time = time.monotonic()
            worker_heartbeat = time.monotonic()
            ctx = await browser.new_context()
            page = await ctx.new_page()

            try:
                output = await asyncio.create_task(
                    jobs_defs[job.name](**job.input).execute(page)
                )
                worker_logger.info(f"Job {job.id} completed successfully")
                await _database.update_job_status(
                    db,
                    worker=name,
                    job_id=job.id,
                    status=_jobs.JobStatus.DONE,
                    processing_time=_calc_processing_time(start_time),
                    output=output,
                )
                job = None
                start_time = None

            except PlaywrightError:
                # We only catch PlaywrightError since they are somewhat expected
                # (e.g. network issues, invalid URLs). Other exceptions like
                # bugs in job implementation should crash the worker to surface
                # the issue.
                worker_logger.exception(
                    f"Playwright error occurred for job {job.id}."
                    " Marking job as failed."
                )
                await _database.update_job_status(
                    db,
                    worker=name,
                    job_id=job.id,
                    status=_jobs.JobStatus.FAILED,
                    processing_time=_calc_processing_time(start_time),
                    output=None,
                )
                job = None
                start_time = None

            finally:
                await page.close()
                await ctx.close()

    except BaseException as e:
        if job:
            worker_logger.info(
                f"Job {job.id} was in progress, marking as failed"
            )
            await _database.update_job_status(
                db,
                worker=name,
                job_id=job.id,
                status=_jobs.JobStatus.FAILED,
                processing_time=(
                    _calc_processing_time(start_time) if start_time else 0
                ),
                output=None,
            )

        # Don't propagate CancelledError since it's expected during shutdown
        if not isinstance(e, asyncio.CancelledError):
            raise

    finally:
        await pw.stop()
        await db.close()


def _calc_processing_time(s: float) -> int:
    # Calculate job processing time in milliseconds
    return round((time.monotonic() - s) * 1000)


def _shutdown(main_task: asyncio.Task, s: signal.Signals) -> None:
    logger.info(f"Received shutdown signal {s.name!r}. Shutting down...")
    main_task.cancel()


def start_worker(name: str, db_path: str, jobs_path: str) -> None:
    loop = asyncio.get_event_loop()
    main_task = loop.create_task(_worker_loop(name, db_path, jobs_path))

    for s in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(s, lambda sig=s: _shutdown(main_task, sig))

    try:
        loop.run_until_complete(main_task)
    except Exception as e:
        # TargetClosedError is expected when stopping worker during job execution
        # since it's caused by browser context being closed. Other Playwright errors
        # that occur during normal job execution are caught in the worker loop.
        if not isinstance(e, TargetClosedError):
            logger.error(
                "Worker process is shutting down due to unexpected error"
            )
            raise
    finally:
        loop.close()



from importlib.metadata import version, PackageNotFoundError

from browsy._jobs import BaseJob, Page

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"

__all__ = ["BaseJob", "Page"]
